// Copyright (C) 2000, Edward W. Felten.  All rights reserved.

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InvalidClassException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;


public class AccountDB implements Serializable {
  private static Object lock = new Object();
  private static List accts = fromFile();
  private static List listeners = new ArrayList();

  private static final String FileName = "bank.db";
    
  public static Account get(int num) {
    synchronized(lock){
      return (Account)(accts.get(num));
    }
  }

  public static int size() {
    synchronized(lock){
      return accts.size();
    }
  }

  public static void add(Account a) {
    synchronized(lock){
      int sz = accts.size();
      a.setNum(sz);
      accts.add(a);
      fireTableRowsInserted(sz, sz);
    }
  }

  public static int indexOf(Account a) {
    synchronized(lock) {
      return accts.indexOf(a);
    }
  }

  public static Collection getAccounts() {
    synchronized(lock) {
      return accts;
    }
  }

  public static Iterator getIterator() {
    synchronized(lock) {
      return accts.iterator();
    }
  }

  public static void toFile() throws IOException {
    synchronized(lock) {
      toStream(new FileOutputStream(FileName));
    }
  }

  private static void toStream(OutputStream os) throws IOException {
    ObjectOutputStream oos = new ObjectOutputStream(os);
    oos.writeObject(accts);
    oos.close();
  }

  private static List fromFile() {
    try {
      return fromStream(new FileInputStream(FileName));
    }catch(IOException x){
      return null;
    }
  }

  private static List fromStream(InputStream is) throws IOException {
    try{
      ObjectInputStream ois = new ObjectInputStream(is);
      List ret = (List)(ois.readObject());
      ois.close();
      return ret;
    }catch(InvalidClassException x){
      return null;
    }catch(ClassNotFoundException x){
      System.err.println(x);
      return null;
    }
  }

  public static String toStringVerbose() {
    synchronized(lock) {
      String ret = "";
      Iterator iter = getIterator();
      while(iter.hasNext()){
	Account a = (Account)(iter.next());
	ret = ret+"\n"+a;
      }
      return ret;
    }
  }

  private static void tryToSave() {
    try{
      toFile();
    }catch(IOException x){
      System.err.println("Couldn't save AcountDB --- " + x);
    }
  }

  public static void init() throws IOException {
    synchronized(lock) {
      accts = new ArrayList();
      Account a = new Account("Alice", 0, new Money(100,0));
      Account b = new Account("Bob", 1, new Money(50,0));
      toFile();
    }
  }

  public static void addListener(AcctDataModel adm) {
    synchronized(lock) {
      listeners.add(adm);
    }
  }

  public static void removeListener(AcctDataModel adm) {
    synchronized(lock) {
      listeners.remove(adm);
    }
  }

  private static void fireTableRowsInserted(int firstRow, int lastRow) {
    synchronized(lock) {
      Iterator iter = listeners.listIterator();
      while(iter.hasNext()){
	AcctDataModel adm = (AcctDataModel)(iter.next());
	adm.fireTableRowsInserted(firstRow, lastRow);
      }
      tryToSave();
    }
  }

  public static void fireBalanceChanged(Account a) {
    synchronized(lock) {
      int idx = indexOf(a);
      Iterator iter = listeners.listIterator();
      while(iter.hasNext()){
	AcctDataModel adm = (AcctDataModel)(iter.next());
	adm.fireTableCellUpdated(idx, 2);
      }
      tryToSave();
    }
  }
}
